pub struct Constructed(/* private fields */);Expand description
The extra memory of a class that has been constructed.
This is unsized, and contains every subclass of a given class.
pub struct Constructed(/* private fields */);The extra memory of a class that has been constructed.
This is unsized, and contains every subclass of a given class.